type op = 
	Add    | Sub | Mult    | Div | Equal | Neq 
  | Less   | Leq | Greater | Geq | And   | Or 
  | Concat | Mod | Neg

type expr =
    Literal of int
  | SLiteral of string
  | BLiteral of bool
  | ALiteral of expr list
  | ALitRpt of expr * expr
  | Id of string
  | ArrId of string * expr
  | Binop of expr * op * expr
  | Assign of string * expr * expr
  | Que of string list
  | Call of string * expr list
  | Noexpr

type var_decl =
	VarDecl of string * expr
  |	ArrDecl of expr * string * expr

type stmt =
    Block of stmt list
  | Expr of expr
  | If of expr * stmt * stmt
  | Repeat of expr * expr * stmt
  | Decls of var_decl list
  | Return of expr

type t =
   Void
  | Int
  | Array of t
  | Boolean
  | String
  | Undefined

type named_block_type =
  | Global
  | Question
  | Execute
  | Function

type named_block = {
	name : string;
	formals : string list;
	decls : var_decl list;
	body : stmt list;
	t: named_block_type;
  }

type learn_program = named_block * named_block list * named_block list * named_block

(* string of type *) 						
let rec string_of_type t = match t with
  | Void -> "Void"
  | Int -> "Integer"
  | Array(t') -> "Array of " ^ string_of_type t'
  | Boolean -> "Boolean"
  | String -> "String"
  | Undefined -> "Undefined"

let rec string_of_expr = function
    Literal(l) -> string_of_int l
  | SLiteral(s) -> "\"" ^ s ^ "\""
  | BLiteral(b) -> string_of_bool b
  | ALiteral(a) -> 
	  "[" ^
	  String.concat "," (List.map (fun e -> string_of_expr e) a)
	  ^ "]"
  | ALitRpt(e1, e2) -> ""
  | Id(s) -> s
  | ArrId(s, e) -> s ^ "[" ^ string_of_expr e ^ "]"
  | Binop(e1, o, e2) ->
      string_of_expr e1 ^ " " ^
      (match o with
		Add -> "+"     | Sub -> "-"  | Mult -> "*" | Div -> "/"
      | Equal -> "=="  | Neq -> "!=" | Less -> "<" | Leq -> "<="
      | Greater -> ">" | Geq -> ">=" | Neg -> "!"  | Mod -> "%"
      | Concat -> "^"  | Or -> "or"  | And -> "and") ^ " " ^
      string_of_expr e2
  | Assign(v, e1, e2) -> v ^ string_of_expr_opt "[" "]" e1 ^ "=" ^ string_of_expr e2
	| Que(qlist) -> 
			"-> {" ^
			String.concat "," (List.map (fun id -> id) qlist)
			^ "}"
	| Call(id, el) ->
			id ^ "(" ^
			String.concat "," (List.map string_of_expr el)
			^ ")"
  | Noexpr -> ""

and string_of_expr_opt l r e = match e with
	Noexpr -> string_of_expr e
  |	_ -> l ^ string_of_expr e ^ r

let rec string_of_stmt = function
    Block(stmts) ->
      "{\n" ^ String.concat "" (List.map string_of_stmt stmts) ^ "}\n"
  | Expr(expr) -> string_of_expr expr ^ ";\n";
  | If(e, s, Block([])) -> "if (" ^ string_of_expr e ^ ")\n" ^ string_of_stmt s
  | If(e, s1, s2) ->  "if (" ^ string_of_expr e ^ ")\n" ^
      string_of_stmt s1 ^ "else\n" ^ string_of_stmt s2
  | Repeat(e1, e2, s1) ->
	  	"repeat (" ^ string_of_expr e1 ^ ";" ^ string_of_expr e2 ^ ")" ^ string_of_stmt s1
  | Decls dl ->
			let string_of_vdecl = function
				VarDecl(id, expr) -> "var " ^ id ^ " = " ^ string_of_expr expr ^ ";\n"
			  |	ArrDecl(expr1, id, expr2) -> 
						"var [" ^ string_of_expr expr1 ^ "] " ^ id ^ " = " ^ string_of_expr expr2 ^ ";\n" in
				String.concat "" (List.map string_of_vdecl dl)
  | Return expr -> "return " ^ string_of_expr expr ^ ";\n"
				
let string_of_named_block_type = function
  | Global -> ""
  | Question -> "question"
  | Execute -> "execute"
  | Function -> "function"
				
let string_of_named_block nb = 
	string_of_named_block_type nb.t ^ " " ^ nb.name ^
	(match nb.t with
  | Global -> 
				string_of_stmt (Decls(nb.decls))
  | Question | Execute -> 
				"\n{\n" ^
				string_of_stmt (Decls(nb.decls)) ^
				String.concat "" (List.map string_of_stmt nb.body) ^
				"}\n"
  | Function ->
				"(" ^ String.concat "," nb.formals ^ ")" ^
				"\n{\n" ^
				string_of_stmt (Decls(nb.decls)) ^
				String.concat "" (List.map string_of_stmt nb.body) ^
				"}\n")

let string_of_learn_program (vars, func, ques, exec) =
	string_of_named_block vars ^ "\n" ^
	String.concat "" (List.map string_of_named_block func) ^
	String.concat "" (List.map string_of_named_block ques) ^
	string_of_named_block exec ^ "\n"
