%{  open Ast
		open Array
		open Lexing
		exception ParseFailure of string
		(* Obtain line number where parsing error occurred *)
		let string_of_position s =
				let line_str = string_of_int (s.pos_lnum) in
					"line:" ^ line_str
%}

%token SEMI LPAREN RPAREN LBRACE RBRACE
%token LBLOCK RBLOCK COMMA
%token PLUS MINUS TIMES DIVIDE MOD ASSIGN
%token EQ NEQ LT LEQ GT GEQ
%token NEG CONCAT TRAN VAR
%token IF ELSE ELSEIF REPEAT INT
%token AND OR QUE EXC FUNC RET
%token <int> LITERAL
%token <string> ID
%token <string> STRING
%token <bool> BOOLEAN
%token EOF

%nonassoc NOELSE
%nonassoc ELSE
%nonassoc ELSEIF
%right ASSIGN TRAN
%left OR AND
%left EQ NEQ
%left LT GT LEQ GEQ
%left CONCAT
%left PLUS MINUS
%left TIMES DIVIDE MOD
%right NEG

%start learn_program
%type <Ast.learn_program> learn_program

%%

learn_program:
 	 globals_opt function_decl_list_opt question_decl_list execute_decl { $1, $2, $3, $4 }
 | error { raise (ParseFailure(string_of_position (symbol_start_pos ()))) }
	
globals_opt:
		var_decl_list_opt 
			{ { name = "";
				formals = [];
				decls = List.rev $1;
				body = [];
				t = Global } }

var_decl_list_opt:
	/* nothing */ { [] }
 |	var_decl_list_opt var_decl SEMI { $2 :: $1 }

var_decl:
	VAR ID ASSIGN expr { VarDecl($2, $4) }
 |	VAR block_expr ID ASSIGN expr { ArrDecl($2, $3, $5) }

function_decl_list_opt:
   /* nothing */ { [] }
 | function_decl_list_opt function_decl { $2 :: $1 }

function_decl:
  FUNC ID LPAREN id_list_opt RPAREN LBRACE var_decl_list_opt stmt_list RBRACE
	  { {	name = $2;
			formals = List.rev $4;
			decls = List.rev $7;
			body = List.rev $8;
			t = Function } }

question_decl_list:
	question_decl { $1 :: [] }
 |	question_decl_list question_decl { $2 :: $1 }

question_decl:
	QUE ID LBRACE var_decl_list_opt stmt_list RBRACE
		{ {	name = $2;
			formals = [];
			decls = List.rev $4;
			body = List.rev $5;
			t = Question } }

execute_decl:
	EXC LBRACE var_decl_list_opt stmt_list RBRACE
		{ { name = "";
			formals = [];
			decls = List.rev $3;
			body = List.rev $4;
			t = Execute } }

stmt_list_opt:
	/* nothing */ { [] }
 |	stmt_list { $1 }

stmt_list:
	stmt { $1 :: [] }
 |	stmt_list stmt { $2 :: $1 }

elif_list:
 	  ELSEIF LPAREN expr RPAREN stmt %prec NOELSE { If($3, $5, Block([])) }
	| ELSEIF LPAREN expr RPAREN stmt ELSE stmt { If($3, $5, $7) }
	| ELSEIF LPAREN expr RPAREN stmt elif_list { If($3, $5, Block([$6])) }
	
stmt:
	expr SEMI { Expr($1) }
 |	LBRACE stmt_list_opt RBRACE { Block(List.rev $2) }
 |	IF LPAREN expr RPAREN stmt %prec NOELSE { If($3, $5, Block([])) }
 |	IF LPAREN expr RPAREN stmt ELSE stmt { If($3, $5, $7) }
 |  IF LPAREN expr RPAREN stmt elif_list { If($3, $5, $6) }
 |	REPEAT LPAREN expr SEMI expr_opt RPAREN stmt { Repeat($3, $5, $7) }
 |  RET expr SEMI { Return($2) }

block_expr_opt:
	/* nothing */ { Noexpr }
 |	block_expr    { $1 }

block_expr:
	LBLOCK expr RBLOCK { $2 }

expr_opt:
	/* nothing */ { Noexpr }
 |	expr          { $1 }

expr_list:
	 expr { $1 :: [] }
 | expr_list COMMA expr { $3 :: $1 }

expr_list_opt:
	 /* nothing */ { [] }
 | expr_list { $1 }

id_list:
	 ID { $1 :: [] }
 | id_list COMMA ID { $3 :: $1 }

id_list_opt:
   /* nothing */ { [] }
 | id_list { $1 }

expr:
	LITERAL          { Literal($1) }
 |  STRING           { SLiteral($1) }
 |  BOOLEAN          { BLiteral($1) }
 |	LBRACE expr_list RBRACE { ALiteral(List.rev $2) }
 |	ID               { Id($1) }
 |  ID block_expr    { ArrId($1, $2) }
 |  MINUS  expr		 { Binop(Literal(-1), Mult,  $2) }
 |  NEG    expr      { Binop(Literal(-1), Neg,   $2) }
 |	expr PLUS   expr { Binop($1, Add,   $3) }
 |	expr MINUS  expr { Binop($1, Sub,   $3) }
 |	expr TIMES  expr { Binop($1, Mult,  $3) }
 |	expr DIVIDE expr { Binop($1, Div,   $3) }
 |  expr MOD    expr { Binop($1, Mod,   $3) }
 |  expr CONCAT expr { Binop($1, Concat, $3) }
 |	expr EQ     expr { Binop($1, Equal, $3) }
 |	expr NEQ    expr { Binop($1, Neq,   $3) }
 |	expr LT     expr { Binop($1, Less,  $3) }
 |	expr LEQ    expr { Binop($1, Leq,   $3) }
 |	expr GT     expr { Binop($1, Greater, $3) }
 |	expr GEQ    expr { Binop($1, Geq,   $3) }
 |  expr OR     expr { Binop($1, Or,    $3) }
 |  expr AND    expr { Binop($1, And,   $3) }
 |  ID block_expr_opt ASSIGN expr { Assign($1, $2, $4) }
 |	TRAN ID          { Que($2 :: []) }
 |  TRAN LBRACE id_list RBRACE 		{ Que($3) }
 |  ID LPAREN expr_list_opt RPAREN 	{ Call($1, List.rev $3) }
 |	LPAREN expr RPAREN { $2 }
